from functools import partial
from uuid import uuid4

from dash_html_components import Div

from . import CustomElement


CustomDiv = CustomElement(Div)

Row = partial(CustomDiv, className="row valign-wrapper")


def Col(s=None, m=None, l=None):
    class_name = f"col"
    if not (s or m or l):
        s = 12

    if s:
        class_name += f" s{s}"

    if m:
        class_name += f" m{m}"

    if l:
        class_name += f" l{l}"

    return partial(CustomDiv, className=class_name)


def Container(*args, **kwargs):
    def wrapper():
        return CustomDiv(*args, **kwargs, className="container", title=str(uuid4()))

    return wrapper



from copy import deepcopy
from typing import Dict

from dash import Dash
from dash_html_components import H5, H6
from dash_core_components import Markdown 
from markdown2 import markdown

from ..components import ComponentGroup
from ..html_helpers.divs import Container, Row
from . import as_component


class KWSDashApp:
    def __init__(self, name="Default Dash App", doc=""):
        self.name = name
        self.components = []
        self.component_layouts = []
        self.callbacks = set()
        self.doc = doc
        self.md_doc = markdown(doc)
        self.registered_components = set()
        self.create_header()

    def create_header(self):
        self.component_layouts.append(
            as_component(H5, "app-title", self.name, container=Row).layout()
        )

        self.component_layouts.append(
            as_component(
                Markdown, "app-description", self.doc, container=Row
            ).layout()
        )

    def register_component(self, component):
        if component.id not in self.registered_components:
            self.components.append(component)

    def _register_component(self, group_or_component, app, top=True):
        self.registered_components.add(group_or_component.id)
        group_or_component._before_registry(app)
        if top:
            self.component_layouts.append(group_or_component.layout())

        for callback in group_or_component.callbacks:
            self.callbacks.add(callback)

        if isinstance(group_or_component, ComponentGroup):
            for component in group_or_component.components:
                self._register_component(component, app, top=False)

    def register_callback(self, callback, outputs=None, inputs=None, states=None):
        callback.register(outputs, inputs, states)
        self.callbacks.add(callback)
        return callback

    def instanciate(self, *args, **kwargs):
        app = Dash(
            __name__,
            *args,
            **kwargs,
            suppress_callback_exceptions=True,
            external_stylesheets=[
                "https://cdnjs.cloudflare.com/ajax/libs/materialize/1.0.0/css/materialize.min.css"
            ],
            external_scripts=[
                "https://cdnjs.cloudflare.com/ajax/libs/materialize/1.0.0/js/materialize.min.js",
            ],
        )

        for component in self.components:
            self._register_component(component, app)

        for callback in self.callbacks:
            if callback.dash_outputs:

                app.callback(
                    callback.dash_outputs, callback.dash_inputs, callback.dash_states
                )(callback.run)

        app.layout = Container(self.component_layouts, id=f"main-div")

        return app

    def run_server(self, *args, **kwargs):
        return self.instanciate().run_server(*args, **kwargs)



import base64
import io
import os

from dash_html_components import Div
from dash_core_components import Upload
import pandas as pd

from dash_infra import ComponentGroup, as_callback, as_component
from dash_infra.components.storage import ContextStorage, PickleStorage


class FileInput(ComponentGroup):
    def __init__(self, id, *parsers, store=None, **kwargs):
        self.id = id
        self.store = store or ContextStorage(f"{id}-store")
        self.parse_to_store = as_callback(
            inputs=(id, "contents"), states=[(id, "filename"), (id, "last_modified")],
        )(self._parse_to_store).to(self.store)
        self.input = as_component(
            Upload, id=id, children=Div("Select a file"), className="file-input row",
        )
        self.parsers = parsers
        if store:
            super().__init__(f"{id}-container", self.input, **kwargs)
        else:
            super().__init__(f"{id}-container", self.input, self.store, **kwargs)

        self.parse_to_store.func.__name__ = self.id

    def _parse_to_store(self, contents, filename, _):
        if not contents:
            return ""
        _, content_string = contents.split(",")
        decoded = base64.b64decode(content_string)
        extension = os.path.splitext(filename)[1].strip(".")
        for parser in self.parsers:
            func = getattr(parser, extension, None)
            if func:
                return func(decoded)

        return ""


class DataFrameParsers(object):
    @staticmethod
    def csv(contents):
        return pd.read_csv(io.StringIO(contents.decode()), sep=";")

    @staticmethod
    def xls(contents):
        return pd.read_excel(io.BytesIO(contents))



from abc import ABC
from copy import deepcopy

from dash.dependencies import Output
from dash_html_components import Div

from ..html_helpers.divs import Row
from .callback import Callback


class Component(object):
    def __init__(self, id):
        self.id = id
        self.callbacks = set()

    def find_callbacks(self):
        callbacks = set()
        for attr_name in dir(self):
            attr = getattr(self, attr_name)

            if isinstance(attr, Callback):
                attr._format(**self.__dict__)
                self.__setattr__(attr.name, attr)
                callbacks.add(attr)

        return callbacks

    def layout(self):
        raise NotImplementedError

    @classmethod
    def from_config(cls, config):
        id = config.get("id")
        if id is None:
            raise AttributeError("No id provided in config")
        component = cls(id)
        for key, value in config.items():
            component.__setattr__(key, value)

        return component

    def _before_registry(self, app):
        self.callbacks = self.find_callbacks()

    def register_callback(self, callback):
        self.callbacks.add(callback)
        return callback


def as_component(dash_object, id, *args, container=None, **kwargs):
    class Wrapper(Component):
        def layout(self):
            if container:
                return container([dash_object(*args, id=id, **kwargs)])
            else:
                return dash_object(*args, id=id, **kwargs)

    return Wrapper(id)



from dash_core_components import Graph
from dash_infra.core import as_callback, as_component

from .store_user import StoreUser


class Figure(StoreUser):
    """
    A Figure component uses a Storage Component as a temporary
    output for all registered callback inputs.
    Each output is stored as a key in the state of the output
    and can be accessed as a dataset.
    """

    def __init__(self, id, store=None, layout=None, **kwargs):
        super().__init__(
            id,
            display_component=Graph,
            display_properties="figure",
            store=store,
            **kwargs,
        )
        self.graph_layout = layout

    def display(self, data):
        graph = {
            "data": data,
            "layout": self.graph_layout,
        }
        return graph


class PandasFigure(Figure):
    def __init__(self, id, dataset_config, types=None, names=None, **kwargs):
        super().__init__(id, **kwargs)
        self.dataset_config = dataset_config

    def _get_col(self, df, label):
        if isinstance(label, int):
            return df.columns[label]
        return label

    def _process_data(self, data):
        datasets = []
        for dataset_name, config in self.dataset_config.items():
            dataset = data[dataset_name]
            x_cols = config.get("x_cols", [])
            y_cols = config.get("y_cols", [])
            types = config.get("types", ["markers"] * len(x_cols))
            names = config.get(
                "names",
                [
                    f"{self._get_col(dataset, x)} vs {self._get_col(dataset, y)}"
                    for (x, y) in zip(x_cols, y_cols)
                ],
            )

            datasets += [
                {
                    "x": list(dataset[self._get_col(dataset, x)]),
                    "y": list(dataset[self._get_col(dataset, y)]),
                    "mode": t,
                    "name": name,
                }
                for x, y, t, name in zip(x_cols, y_cols, types, names)
            ]

        return datasets



from dash_infra.core import Component
from dash_infra.html_helpers.divs import Row


class ComponentGroup(Component):
    def __init__(self, id, *components, container=Row):
        super().__init__(id)
        self.components = components
        self.container = container

    def layout(self):
        return self.container([c.layout() for c in self.components], id=self.id)



from dash.dependencies import Output
from dash_infra.components.storage import PickleStorage, ContextStorage
from dash_infra.core import as_callback, as_component
from dash_infra.html_helpers.divs import Col, Row

from .groups import ComponentGroup


class StoreUser(ComponentGroup):
    """
    A Store User component is a component which features a store,
    and uses that store as the input for its display.
    """

    registered_stores = set()

    def __init__(
        self,
        id,
        display_component,
        display_properties,
        store=None,
        keys=None,
        container=Row,
        **kwargs,
    ):
        self.display_component = as_component(
            display_component, id, container=Col(), **kwargs
        )
        self.storage = store or ContextStorage(id=f"store-{id}")
        self.display_id = id
        self.display_properties = display_properties
        self.keys = keys or []
        self.registered_stores.add(store.id)
        if store:
            super().__init__(
                f"{id}-container", self.display_component, container=container
            )
        else:
            super().__init__(
                f"{id}-container", self.display_component, store, container=container
            )

        self.display_callback = self.create_display_callback()
        self.display_callback.use_storage_inputs(self.storage, *self.keys)

    def create_display_callback(self):
        if isinstance(self.display_properties, list):
            return as_callback(
                outputs=[(self.display_id, p) for p in self.display_properties]
            )(self._display_callback)
        else:
            return as_callback(outputs=(self.display_id, self.display_properties))(
                self._display_callback
            )

    def register_callback(self, callback):
        """
        Registers a callback
        """
        return callback.to(self.storage)

    def _before_registry(self, app):
        self.callbacks.add(self.display_callback)

    def _process_data(self, data):
        return data

    def _display_callback(self, data):
        data = self._process_data(data)
        return self.display(data)

    def display(self, data):
        raise NotImplementedError
